==> mapillary-js-2.1.0.jar: META-INF/resources/webjars/mapillary-js/2.1.0/src/render/DOMRenderer.ts <==
/// <reference path="../../typings/index.d.ts" />

import * as _ from "underscore";
import * as vd from "virtual-dom";

import {Observable} from "rxjs/Observable";
import {Subject} from "rxjs/Subject";

import "rxjs/add/operator/combineLatest";
import "rxjs/add/operator/distinctUntilChanged";
import "rxjs/add/operator/filter";
import "rxjs/add/operator/map";
import "rxjs/add/operator/pluck";
import "rxjs/add/operator/scan";

import {ISize, IVNodeHash, RenderMode, RenderService} from "../Render";
import {IFrame} from "../State";

interface INodePatch {
    vnode: vd.VNode;
    vpatch: vd.VPatch[];
}

interface IVNodeHashes {
    [name: string]: vd.VNode;
}

interface IOffset {
    bottom: number;
    left: number;
    right: number;
    top: number;
}

interface IAdaptive {
    elementHeight: number;
    elementWidth: number;
    imageAspect: number;
    renderMode: RenderMode;
}

interface IAdaptiveOperation {
    (adaptive: IAdaptive): IAdaptive;
}

export class DOMRenderer {
    private _renderService: RenderService;
    private _currentFrame$: Observable<IFrame>;

    private _adaptiveOperation$: Subject<IAdaptiveOperation> = new Subject<IAdaptiveOperation>();
    private _offset$: Observable<IOffset>;

    private _element$: Observable<Element>;
    private _vPatch$: Observable<vd.VPatch[]>;
    private _vNode$: Observable<vd.VNode>;
    private _render$: Subject<IVNodeHash> = new Subject<IVNodeHash>();
    private _renderAdaptive$: Subject<IVNodeHash> = new Subject<IVNodeHash>();

    constructor (element: HTMLElement, renderService: RenderService, currentFrame$: Observable<IFrame>) {
        this._renderService = renderService;
        this._currentFrame$ = currentFrame$;

        let rootNode: Element = vd.create(vd.h("div.domRenderer", []));
        element.appendChild(rootNode);

        this._offset$ = this._adaptiveOperation$
            .scan<IAdaptive>(
                (adaptive: IAdaptive, operation: IAdaptiveOperation): IAdaptive => {
                    return operation(adaptive);
                },
                {
                    elementHeight: element.offsetHeight,
                    elementWidth: element.offsetWidth,
                    imageAspect: 0,
                    renderMode: RenderMode.Fill,
                })
            .filter(
                (adaptive: IAdaptive): boolean => {
                    return adaptive.imageAspect > 0 && adaptive.elementWidth > 0 && adaptive.elementHeight > 0;
                })
            .map<IOffset>(
                (adaptive: IAdaptive): IOffset => {
                    let elementAspect: number = adaptive.elementWidth / adaptive.elementHeight;
                    let ratio: number = adaptive.imageAspect / elementAspect;

                    let verticalOffset: number = 0;
                    let horizontalOffset: number = 0;

                    if (adaptive.renderMode === RenderMode.Letterbox) {
                        if (adaptive.imageAspect > elementAspect) {
                            verticalOffset = adaptive.elementHeight * (1 - 1 / ratio) / 2;
                        } else {
                            horizontalOffset = adaptive.elementWidth * (1 - ratio) / 2;
                        }
                    } else {
                        if (adaptive.imageAspect > elementAspect) {
                            horizontalOffset = -adaptive.elementWidth * (ratio - 1) / 2;
                        } else {
                            verticalOffset = -adaptive.elementHeight * (1 / ratio - 1) / 2;
                        }
                    }

                    return {
                        bottom: verticalOffset,
                        left: horizontalOffset,
                        right: horizontalOffset,
                        top: verticalOffset,
                    };
                });

        this._currentFrame$
            .filter(
                (frame: IFrame): boolean => {
                    return frame.state.currentNode != null;
                })
            .distinctUntilChanged(
                (k1: string, k2: string): boolean => {
                    return k1 === k2;
                },
                (frame: IFrame): string => {
                    return frame.state.currentNode.key;
                })
            .map<number>(
                (frame: IFrame): number => {
                    return frame.state.currentTransform.basicAspect;
                })
            .map<IAdaptiveOperation>(
                 (aspect: number): IAdaptiveOperation => {
                    return (adaptive: IAdaptive): IAdaptive => {
                        adaptive.imageAspect = aspect;

                        return adaptive;
                    };
                })
            .subscribe(this._adaptiveOperation$);

        this._renderAdaptive$
            .scan<IVNodeHashes>(
                (vNodeHashes: IVNodeHashes, vNodeHash: IVNodeHash): IVNodeHashes => {
                    if (vNodeHash.vnode == null) {
                        delete vNodeHashes[vNodeHash.name];
                    } else {
                        vNodeHashes[vNodeHash.name] = vNodeHash.vnode;
                    }
                    return vNodeHashes;
                },
                {})
            .combineLatest(this._offset$)
            .map<IVNodeHash>(
                (vo: [IVNodeHashes, IOffset]): IVNodeHash => {
                    let vNodes: vd.VNode[] = _.values(vo[0]);
                    let offset: IOffset = vo[1];

                    let properties: vd.createProperties = {
                        style: {
                            bottom: offset.bottom + "px",
                            left: offset.left + "px",
                            position: "absolute",
                            right: offset.right + "px",
                            top: offset.top + "px",
                            zIndex: -1,
                        },
                    };

                    return {
                        name: "adaptiveDomRenderer",
                        vnode: vd.h("div.adaptiveDomRenderer", properties, vNodes),
                    };
                })
            .subscribe(this._render$);

        this._vNode$ = this._render$
            .scan<IVNodeHashes>(
                (vNodeHashes: IVNodeHashes, vNodeHash: IVNodeHash): IVNodeHashes => {
                    if (vNodeHash.vnode == null) {
                        delete vNodeHashes[vNodeHash.name];
                    } else {
                        vNodeHashes[vNodeHash.name] = vNodeHash.vnode;
                    }

                    return vNodeHashes;
                },
                {})
            .map(
                (vNodeHashes: IVNodeHashes): vd.VNode => {
                    let vNodes: vd.VNode[] = _.values(vNodeHashes);
                    return vd.h("div.domRenderer", vNodes);
                });

        this._vPatch$ = this._vNode$
            .scan<INodePatch>(
                (nodePatch: INodePatch, vNode: vd.VNode): INodePatch => {
                    nodePatch.vpatch = vd.diff(nodePatch.vnode, vNode);
                    nodePatch.vnode = vNode;
                    return nodePatch;
                },
                {vnode: vd.h("div.domRenderer", []), vpatch: null})
            .pluck<vd.VPatch[]>("vpatch");

        this._element$ = this._vPatch$
            .scan(
                (oldElement: Element, vPatch: vd.VPatch[]): Element => {
                    return vd.patch(oldElement, vPatch);
                },
                rootNode)
            .publishReplay(1)
            .refCount();

        this._element$.subscribe();

        this._renderService.size$
            .map<IAdaptiveOperation>(
                (size: ISize): IAdaptiveOperation => {
                    return (adaptive: IAdaptive): IAdaptive => {
                        adaptive.elementWidth = size.width;
                        adaptive.elementHeight = size.height;

                        return adaptive;
                    };
                })
            .subscribe(this._adaptiveOperation$);

        this._renderService.renderMode$
            .map<IAdaptiveOperation>(
                (renderMode: RenderMode): IAdaptiveOperation => {
                    return (adaptive: IAdaptive): IAdaptive => {
                        adaptive.renderMode = renderMode;

                        return adaptive;
                    };
                })
            .subscribe(this._adaptiveOperation$);
    }

    public get element$(): Observable<Element> {
        return this._element$;
    }

    public get render$(): Subject<IVNodeHash> {
        return this._render$;
    }

    public get renderAdaptive$(): Subject<IVNodeHash> {
        return this._renderAdaptive$;
    }

    public clear(name: string): void {
        this._renderAdaptive$.next({name: name, vnode: null});
        this._render$.next({name: name, vnode: null});
    }
}

export default DOMRenderer;

==> mapillary-js-2.3.0.jar: META-INF/resources/webjars/mapillary-js/2.3.0/src/render/DOMRenderer.ts <==
/// <reference path="../../typings/index.d.ts" />

import * as _ from "underscore";
import * as vd from "virtual-dom";

import {Observable} from "rxjs/Observable";
import {Subject} from "rxjs/Subject";

import "rxjs/add/operator/combineLatest";
import "rxjs/add/operator/distinctUntilChanged";
import "rxjs/add/operator/filter";
import "rxjs/add/operator/map";
import "rxjs/add/operator/pluck";
import "rxjs/add/operator/scan";

import {ISize, IVNodeHash, RenderMode, RenderService} from "../Render";
import {IFrame} from "../State";

interface INodePatch {
    vnode: vd.VNode;
    vpatch: vd.VPatch[];
}

interface IVNodeHashes {
    [name: string]: vd.VNode;
}

interface IOffset {
    bottom: number;
    left: number;
    right: number;
    top: number;
}

interface IAdaptive {
    elementHeight: number;
    elementWidth: number;
    imageAspect: number;
    renderMode: RenderMode;
}

interface IAdaptiveOperation {
    (adaptive: IAdaptive): IAdaptive;
}

export class DOMRenderer {
    private _renderService: RenderService;
    private _currentFrame$: Observable<IFrame>;

    private _adaptiveOperation$: Subject<IAdaptiveOperation> = new Subject<IAdaptiveOperation>();
    private _offset$: Observable<IOffset>;

    private _element$: Observable<Element>;
    private _vPatch$: Observable<vd.VPatch[]>;
    private _vNode$: Observable<vd.VNode>;
    private _render$: Subject<IVNodeHash> = new Subject<IVNodeHash>();
    private _renderAdaptive$: Subject<IVNodeHash> = new Subject<IVNodeHash>();

    constructor (element: HTMLElement, renderService: RenderService, currentFrame$: Observable<IFrame>) {
        this._renderService = renderService;
        this._currentFrame$ = currentFrame$;

        let rootNode: Element = vd.create(vd.h("div.domRenderer", []));
        element.appendChild(rootNode);

        this._offset$ = this._adaptiveOperation$
            .scan(
                (adaptive: IAdaptive, operation: IAdaptiveOperation): IAdaptive => {
                    return operation(adaptive);
                },
                {
                    elementHeight: element.offsetHeight,
                    elementWidth: element.offsetWidth,
                    imageAspect: 0,
                    renderMode: RenderMode.Fill,
                })
            .filter(
                (adaptive: IAdaptive): boolean => {
                    return adaptive.imageAspect > 0 && adaptive.elementWidth > 0 && adaptive.elementHeight > 0;
                })
            .map(
                (adaptive: IAdaptive): IOffset => {
                    let elementAspect: number = adaptive.elementWidth / adaptive.elementHeight;
                    let ratio: number = adaptive.imageAspect / elementAspect;

                    let verticalOffset: number = 0;
                    let horizontalOffset: number = 0;

                    if (adaptive.renderMode === RenderMode.Letterbox) {
                        if (adaptive.imageAspect > elementAspect) {
                            verticalOffset = adaptive.elementHeight * (1 - 1 / ratio) / 2;
                        } else {
                            horizontalOffset = adaptive.elementWidth * (1 - ratio) / 2;
                        }
                    } else {
                        if (adaptive.imageAspect > elementAspect) {
                            horizontalOffset = -adaptive.elementWidth * (ratio - 1) / 2;
                        } else {
                            verticalOffset = -adaptive.elementHeight * (1 / ratio - 1) / 2;
                        }
                    }

                    return {
                        bottom: verticalOffset,
                        left: horizontalOffset,
                        right: horizontalOffset,
                        top: verticalOffset,
                    };
                });

        this._currentFrame$
            .filter(
                (frame: IFrame): boolean => {
                    return frame.state.currentNode != null;
                })
            .distinctUntilChanged(
                (k1: string, k2: string): boolean => {
                    return k1 === k2;
                },
                (frame: IFrame): string => {
                    return frame.state.currentNode.key;
                })
            .map(
                (frame: IFrame): number => {
                    return frame.state.currentTransform.basicAspect;
                })
            .map(
                 (aspect: number): IAdaptiveOperation => {
                    return (adaptive: IAdaptive): IAdaptive => {
                        adaptive.imageAspect = aspect;

                        return adaptive;
                    };
                })
            .subscribe(this._adaptiveOperation$);

        this._renderAdaptive$
            .scan(
                (vNodeHashes: IVNodeHashes, vNodeHash: IVNodeHash): IVNodeHashes => {
                    if (vNodeHash.vnode == null) {
                        delete vNodeHashes[vNodeHash.name];
                    } else {
                        vNodeHashes[vNodeHash.name] = vNodeHash.vnode;
                    }
                    return vNodeHashes;
                },
                {})
            .combineLatest(this._offset$)
            .map(
                (vo: [IVNodeHashes, IOffset]): IVNodeHash => {
                    let vNodes: vd.VNode[] = _.values(vo[0]);
                    let offset: IOffset = vo[1];

                    let properties: vd.createProperties = {
                        style: {
                            bottom: offset.bottom + "px",
                            left: offset.left + "px",
                            position: "absolute",
                            right: offset.right + "px",
                            top: offset.top + "px",
                            zIndex: -1,
                        },
                    };

                    return {
                        name: "adaptiveDomRenderer",
                        vnode: vd.h("div.adaptiveDomRenderer", properties, vNodes),
                    };
                })
            .subscribe(this._render$);

        this._vNode$ = this._render$
            .scan(
                (vNodeHashes: IVNodeHashes, vNodeHash: IVNodeHash): IVNodeHashes => {
                    if (vNodeHash.vnode == null) {
                        delete vNodeHashes[vNodeHash.name];
                    } else {
                        vNodeHashes[vNodeHash.name] = vNodeHash.vnode;
                    }

                    return vNodeHashes;
                },
                {})
            .map(
                (vNodeHashes: IVNodeHashes): vd.VNode => {
                    let vNodes: vd.VNode[] = _.values(vNodeHashes);
                    return vd.h("div.domRenderer", vNodes);
                });

        this._vPatch$ = this._vNode$
            .scan(
                (nodePatch: INodePatch, vNode: vd.VNode): INodePatch => {
                    nodePatch.vpatch = vd.diff(nodePatch.vnode, vNode);
                    nodePatch.vnode = vNode;
                    return nodePatch;
                },
                {vnode: vd.h("div.domRenderer", []), vpatch: null})
            .pluck<INodePatch, vd.VPatch[]>("vpatch");

        this._element$ = this._vPatch$
            .scan(
                (oldElement: Element, vPatch: vd.VPatch[]): Element => {
                    return vd.patch(oldElement, vPatch);
                },
                rootNode)
            .publishReplay(1)
            .refCount();

        this._element$.subscribe(() => { /*noop*/ });

        this._renderService.size$
            .map(
                (size: ISize): IAdaptiveOperation => {
                    return (adaptive: IAdaptive): IAdaptive => {
                        adaptive.elementWidth = size.width;
                        adaptive.elementHeight = size.height;

                        return adaptive;
                    };
                })
            .subscribe(this._adaptiveOperation$);

        this._renderService.renderMode$
            .map(
                (renderMode: RenderMode): IAdaptiveOperation => {
                    return (adaptive: IAdaptive): IAdaptive => {
                        adaptive.renderMode = renderMode;

                        return adaptive;
                    };
                })
            .subscribe(this._adaptiveOperation$);
    }

    public get element$(): Observable<Element> {
        return this._element$;
    }

    public get render$(): Subject<IVNodeHash> {
        return this._render$;
    }

    public get renderAdaptive$(): Subject<IVNodeHash> {
        return this._renderAdaptive$;
    }

    public clear(name: string): void {
        this._renderAdaptive$.next({name: name, vnode: null});
        this._render$.next({name: name, vnode: null});
    }
}

export default DOMRenderer;
